package jazzlib;

public class Inflater {
	private static final int[] CPLENS;
	private static final int[] CPLEXT;
	private static final int[] CPDIST;
	private static final int[] CPDEXT;
	private static final int DECODE_HEADER = 0;
	private static final int DECODE_DICT = 1;
	private static final int DECODE_BLOCKS = 2;
	private static final int DECODE_STORED_LEN1 = 3;
	private static final int DECODE_STORED_LEN2 = 4;
	private static final int DECODE_STORED = 5;
	private static final int DECODE_DYN_HEADER = 6;
	private static final int DECODE_HUFFMAN = 7;
	private static final int DECODE_HUFFMAN_LENBITS = 8;
	private static final int DECODE_HUFFMAN_DIST = 9;
	private static final int DECODE_HUFFMAN_DISTBITS = 10;
	private static final int DECODE_CHKSUM = 11;
	private static final int FINISHED = 12;
	private int mode;
	private int readAdler;
	private int neededBits;
	private int repLength;
	private int repDist;
	private int uncomprLen;
	private boolean isLastBlock;
	private int totalOut;
	private int totalIn;
	private boolean nowrap;
	private StreamManipulator input;
	private OutputWindow outputWindow;
	private InflaterDynHeader dynHeader;
	private InflaterHuffmanTree litlenTree;
	private InflaterHuffmanTree distTree;
	private Adler32 adler;

	public Inflater() {
		this(false);
	}

	public Inflater(final boolean nowrap) {
		this.nowrap = nowrap;
		this.adler = new Adler32();
		this.input = new StreamManipulator();
		this.outputWindow = new OutputWindow();
		this.mode = (nowrap ? 2 : 0);
	}

	public void alloc(final int window_SIZE) {
		this.outputWindow.WINDOW_SIZE = window_SIZE;
		this.outputWindow.WINDOW_MASK = window_SIZE - 1;
		this.outputWindow.window = new byte[window_SIZE];
	}

	public void free() {
		this.outputWindow.window = null;
	}

	public void end() {
		this.outputWindow = null;
		this.input = null;
		this.dynHeader = null;
		this.litlenTree = null;
		this.distTree = null;
		this.adler = null;
	}

	public boolean finished() {
		return this.mode == 12 && this.outputWindow.getAvailable() == 0;
	}

	public int getAdler() {
		return this.needsDictionary() ? this.readAdler : ((int) this.adler.getValue());
	}

	public int getRemaining() {
		return this.input.getAvailableBytes();
	}

	public int getTotalIn() {
		return this.totalIn - this.getRemaining();
	}

	public int getTotalOut() {
		return this.totalOut;
	}

	public int inflate(final byte[] array) throws DataFormatException {
		return this.inflate(array, 0, array.length);
	}

	public int inflate(final byte[] array, int n, int n2) throws DataFormatException {
		if (n2 == 0) {
			return 0;
		}
		if (0 > n || n > n + n2 || n + n2 > array.length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		int n3 = 0;
		do {
			if (this.mode != 11) {
				final int copyOutput = this.outputWindow.copyOutput(array, n, n2);
				n += copyOutput;
				n3 += copyOutput;
				this.totalOut += copyOutput;
				n2 -= copyOutput;
				if (n2 == 0) {
					return n3;
				}
				continue;
			}
		} while (this.decode() || (this.outputWindow.getAvailable() > 0 && this.mode != 11));
		return n3;
	}

	public boolean needsDictionary() {
		return this.mode == 1 && this.neededBits == 0;
	}

	public boolean needsInput() {
		return this.input.needsInput();
	}

	public void reset() {
		this.mode = (this.nowrap ? 2 : 0);
		final boolean b = false;
		this.totalOut = (b ? 1 : 0);
		this.totalIn = (b ? 1 : 0);
		this.input.reset();
		this.outputWindow.reset();
		this.dynHeader = null;
		this.litlenTree = null;
		this.distTree = null;
		this.isLastBlock = false;
		this.adler.reset();
	}

	public void setDictionary(final byte[] array) {
		this.setDictionary(array, 0, array.length);
	}

	public void setDictionary(final byte[] array, final int n, final int n2) {
		if (!this.needsDictionary()) {
			throw new IllegalStateException();
		}
		this.outputWindow.copyDict(array, n, n2);
		this.mode = 2;
	}

	public void setInput(final byte[] array) {
		this.setInput(array, 0, array.length);
	}

	public void setInput(final byte[] array, final int n, final int n2) {
		this.input.setInput(array, n, n2);
		this.totalIn += n2;
	}

	private boolean decodeHeader() throws DataFormatException {
		int n;
		if (this.input.bits_in_buffer < 16) {
			if (this.input.window_start == this.input.window_end) {
				n = -1;
			} else {
				final StreamManipulator input = this.input;
				final int buffer = input.buffer;
				final int n2 = (this.input.window[this.input.window_start++] & 0xFF)
						| (this.input.window[this.input.window_start++] & 0xFF) << 8;
				final StreamManipulator input2 = this.input;
				final int bits_in_buffer = input2.bits_in_buffer + 16;
				input2.bits_in_buffer = bits_in_buffer;
				final int buffer2 = buffer | n2 << bits_in_buffer - 16;
				input.buffer = buffer2;
				n = (buffer2 & 0xFFFF);
			}
		} else {
			n = (this.input.buffer & 0xFFFF);
		}
		final int n3 = n;
		if (n3 < 0) {
			return false;
		}
		final StreamManipulator input3 = this.input;
		input3.buffer >>>= 16;
		final StreamManipulator input4 = this.input;
		input4.bits_in_buffer -= 16;
		final int n4 = (n3 << 8 | n3 >> 8) & 0xFFFF;
		if (n4 % 31 != 0) {
			throw new DataFormatException("Header checksum illegal");
		}
		if ((n4 & 0xF00) != 0x800) {
			throw new DataFormatException("Compression Method unknown");
		}
		if ((n4 & 0x20) == 0x0) {
			this.mode = 2;
		} else {
			this.mode = 1;
			this.neededBits = 32;
		}
		return true;
	}

	private boolean decodeDict() {
		while (this.neededBits > 0) {
			int n;
			if (this.input.bits_in_buffer < 8) {
				if (this.input.window_start == this.input.window_end) {
					n = -1;
				} else {
					final StreamManipulator input = this.input;
					final int buffer = input.buffer;
					final int n2 = (this.input.window[this.input.window_start++] & 0xFF)
							| (this.input.window[this.input.window_start++] & 0xFF) << 8;
					final StreamManipulator input2 = this.input;
					final int bits_in_buffer = input2.bits_in_buffer + 16;
					input2.bits_in_buffer = bits_in_buffer;
					final int buffer2 = buffer | n2 << bits_in_buffer - 16;
					input.buffer = buffer2;
					n = (buffer2 & 0xFF);
				}
			} else {
				n = (this.input.buffer & 0xFF);
			}
			final int n3 = n;
			if (n3 < 0) {
				return false;
			}
			final StreamManipulator input3 = this.input;
			input3.buffer >>>= 8;
			final StreamManipulator input4 = this.input;
			input4.bits_in_buffer -= 8;
			this.readAdler = (this.readAdler << 8 | n3);
			this.neededBits -= 8;
		}
		return false;
	}

	private boolean decodeHuffman() throws DataFormatException {
		int i = this.outputWindow.getFreeSpace();
		while (i >= 258) {
			Label_0159: {
				switch (this.mode) {
				case 7: {
					int symbol;
					while (((symbol = this.litlenTree.getSymbol(this.input)) & 0xFFFFFF00) == 0x0) {
						this.outputWindow.write(symbol);
						if (--i < 258) {
							return true;
						}
					}
					if (symbol >= 257) {
						try {
							this.repLength = Inflater.CPLENS[symbol - 257];
							this.neededBits = Inflater.CPLEXT[symbol - 257];
						} catch (ArrayIndexOutOfBoundsException ex) {
							throw new DataFormatException("Illegal rep length code");
						}
						break Label_0159;
					}
					if (symbol < 0) {
						return false;
					}
					this.distTree = null;
					this.litlenTree = null;
					this.mode = 2;
					return true;
				}
				case 8: {
					if (this.neededBits > 0) {
						this.mode = 8;
						int n;
						if (this.input.bits_in_buffer < this.neededBits) {
							if (this.input.window_start == this.input.window_end) {
								n = -1;
							} else {
								final StreamManipulator input = this.input;
								final int buffer = input.buffer;
								final int n2 = (this.input.window[this.input.window_start++] & 0xFF)
										| (this.input.window[this.input.window_start++] & 0xFF) << 8;
								final StreamManipulator input2 = this.input;
								final int bits_in_buffer = input2.bits_in_buffer + 16;
								input2.bits_in_buffer = bits_in_buffer;
								final int buffer2 = buffer | n2 << bits_in_buffer - 16;
								input.buffer = buffer2;
								n = (buffer2 & (1 << this.neededBits) - 1);
							}
						} else {
							n = (this.input.buffer & (1 << this.neededBits) - 1);
						}
						final int n3 = n;
						if (n3 < 0) {
							return false;
						}
						final StreamManipulator input3 = this.input;
						input3.buffer >>>= this.neededBits;
						final StreamManipulator input4 = this.input;
						input4.bits_in_buffer -= this.neededBits;
						this.repLength += n3;
					}
					this.mode = 9;
				}
				case 9: {
					final int symbol2 = this.distTree.getSymbol(this.input);
					if (symbol2 < 0) {
						return false;
					}
					try {
						this.repDist = Inflater.CPDIST[symbol2];
						this.neededBits = Inflater.CPDEXT[symbol2];
					} catch (ArrayIndexOutOfBoundsException ex2) {
						throw new DataFormatException("Illegal rep dist code");
					}
				}
				case 10: {
					if (this.neededBits > 0) {
						this.mode = 10;
						int n4;
						if (this.input.bits_in_buffer < this.neededBits) {
							if (this.input.window_start == this.input.window_end) {
								n4 = -1;
							} else {
								final StreamManipulator input5 = this.input;
								final int buffer3 = input5.buffer;
								final int n5 = (this.input.window[this.input.window_start++] & 0xFF)
										| (this.input.window[this.input.window_start++] & 0xFF) << 8;
								final StreamManipulator input6 = this.input;
								final int bits_in_buffer2 = input6.bits_in_buffer + 16;
								input6.bits_in_buffer = bits_in_buffer2;
								final int buffer4 = buffer3 | n5 << bits_in_buffer2 - 16;
								input5.buffer = buffer4;
								n4 = (buffer4 & (1 << this.neededBits) - 1);
							}
						} else {
							n4 = (this.input.buffer & (1 << this.neededBits) - 1);
						}
						final int n6 = n4;
						if (n6 < 0) {
							return false;
						}
						final StreamManipulator input7 = this.input;
						input7.buffer >>>= this.neededBits;
						final StreamManipulator input8 = this.input;
						input8.bits_in_buffer -= this.neededBits;
						this.repDist += n6;
					}
					final OutputWindow outputWindow = this.outputWindow;
					if ((outputWindow.window_filled += this.repLength) > this.outputWindow.WINDOW_SIZE) {
						throw new IllegalStateException("Window full");
					}
					int n7 = this.outputWindow.window_end - this.repDist & this.outputWindow.WINDOW_MASK;
					final int n8 = this.outputWindow.WINDOW_SIZE - this.repLength;
					if (n7 <= n8 && this.outputWindow.window_end < n8) {
						if (this.repLength <= this.repDist) {
							System.arraycopy(this.outputWindow.window, n7, this.outputWindow.window, this.outputWindow.window_end, this.repLength);
							final OutputWindow outputWindow2 = this.outputWindow;
							outputWindow2.window_end += this.repLength;
						} else {
							int repLength = this.repLength;
							while (repLength-- > 0) {
								this.outputWindow.window[this.outputWindow.window_end++] = this.outputWindow.window[n7++];
							}
						}
					} else {
						this.outputWindow.slowRepeat(n7, this.repLength, this.repDist);
					}
					i -= this.repLength;
					this.mode = 7;
					continue;
				}
				default: {
					throw new IllegalStateException();
				}
				}
			}
		}
		return true;
	}

	private boolean decodeChksum() throws DataFormatException {
		while (this.neededBits > 0) {
			int n;
			if (this.input.bits_in_buffer < 8) {
				if (this.input.window_start == this.input.window_end) {
					n = -1;
				} else {
					final StreamManipulator input = this.input;
					final int buffer = input.buffer;
					final int n2 = (this.input.window[this.input.window_start++] & 0xFF)
							| (this.input.window[this.input.window_start++] & 0xFF) << 8;
					final StreamManipulator input2 = this.input;
					final int bits_in_buffer = input2.bits_in_buffer + 16;
					input2.bits_in_buffer = bits_in_buffer;
					final int buffer2 = buffer | n2 << bits_in_buffer - 16;
					input.buffer = buffer2;
					n = (buffer2 & 0xFF);
				}
			} else {
				n = (this.input.buffer & 0xFF);
			}
			final int n3 = n;
			if (n3 < 0) {
				return false;
			}
			final StreamManipulator input3 = this.input;
			input3.buffer >>>= 8;
			final StreamManipulator input4 = this.input;
			input4.bits_in_buffer -= 8;
			this.readAdler = (this.readAdler << 8 | n3);
			this.neededBits -= 8;
		}
		if ((int) this.adler.getValue() != this.readAdler) {
			throw new DataFormatException("Adler chksum doesn't match: "
					+ Integer.toHexString((int) this.adler.getValue()) + " vs. " + Integer.toHexString(this.readAdler));
		}
		this.mode = 12;
		return false;
	}

	private boolean decode() throws DataFormatException {
		switch (this.mode) {
		case 0: {
			return this.decodeHeader();
		}
		case 1: {
			return this.decodeDict();
		}
		case 11: {
			return this.decodeChksum();
		}
		case 2: {
			if (this.isLastBlock) {
				if (this.nowrap) {
					this.mode = 12;
					return false;
				}
				this.input.skipToByteBoundary();
				this.neededBits = 32;
				this.mode = 11;
				return true;
			} else {
				int n;
				if (this.input.bits_in_buffer < 3) {
					if (this.input.window_start == this.input.window_end) {
						n = -1;
					} else {
						final StreamManipulator input = this.input;
						final int buffer = input.buffer;
						final int n2 = (this.input.window[this.input.window_start++] & 0xFF)
								| (this.input.window[this.input.window_start++] & 0xFF) << 8;
						final StreamManipulator input2 = this.input;
						final int bits_in_buffer = input2.bits_in_buffer + 16;
						input2.bits_in_buffer = bits_in_buffer;
						final int buffer2 = buffer | n2 << bits_in_buffer - 16;
						input.buffer = buffer2;
						n = (buffer2 & 0x7);
					}
				} else {
					n = (this.input.buffer & 0x7);
				}
				final int n3 = n;
				if (n3 < 0) {
					return false;
				}
				final StreamManipulator input3 = this.input;
				input3.buffer >>>= 3;
				final StreamManipulator input4 = this.input;
				input4.bits_in_buffer -= 3;
				if ((n3 & 0x1) != 0x0) {
					this.isLastBlock = true;
				}
				switch (n3 >> 1) {
				case 0: {
					this.input.skipToByteBoundary();
					this.mode = 3;
					break;
				}
				case 1: {
					this.litlenTree = InflaterHuffmanTree.defLitLenTree;
					this.distTree = InflaterHuffmanTree.defDistTree;
					this.mode = 7;
					break;
				}
				case 2: {
					this.dynHeader = new InflaterDynHeader();
					this.mode = 6;
					break;
				}
				default: {
					throw new DataFormatException("Unknown block type " + n3);
				}
				}
				return true;
			}
		}
		case 3: {
			int uncomprLen;
			int n4;
			if (this.input.bits_in_buffer < 16) {
				if (this.input.window_start == this.input.window_end) {
					n4 = (uncomprLen = -1);
				} else {
					final StreamManipulator input5 = this.input;
					final int buffer3 = input5.buffer;
					final int n5 = (this.input.window[this.input.window_start++] & 0xFF)
							| (this.input.window[this.input.window_start++] & 0xFF) << 8;
					final StreamManipulator input6 = this.input;
					final int bits_in_buffer2 = input6.bits_in_buffer + 16;
					input6.bits_in_buffer = bits_in_buffer2;
					final int buffer4 = buffer3 | n5 << bits_in_buffer2 - 16;
					input5.buffer = buffer4;
					n4 = (uncomprLen = (buffer4 & 0xFFFF));
				}
			} else {
				n4 = (uncomprLen = (this.input.buffer & 0xFFFF));
			}
			this.uncomprLen = uncomprLen;
			if (n4 < 0) {
				return false;
			}
			final StreamManipulator input7 = this.input;
			input7.buffer >>>= 16;
			final StreamManipulator input8 = this.input;
			input8.bits_in_buffer -= 16;
			this.mode = 4;
		}
		case 4: {
			int n6;
			if (this.input.bits_in_buffer < 16) {
				if (this.input.window_start == this.input.window_end) {
					n6 = -1;
				} else {
					final StreamManipulator input9 = this.input;
					final int buffer5 = input9.buffer;
					final int n7 = (this.input.window[this.input.window_start++] & 0xFF)
							| (this.input.window[this.input.window_start++] & 0xFF) << 8;
					final StreamManipulator input10 = this.input;
					final int bits_in_buffer3 = input10.bits_in_buffer + 16;
					input10.bits_in_buffer = bits_in_buffer3;
					final int buffer6 = buffer5 | n7 << bits_in_buffer3 - 16;
					input9.buffer = buffer6;
					n6 = (buffer6 & 0xFFFF);
				}
			} else {
				n6 = (this.input.buffer & 0xFFFF);
			}
			final int n8 = n6;
			if (n8 < 0) {
				return false;
			}
			final StreamManipulator input11 = this.input;
			input11.buffer >>>= 16;
			final StreamManipulator input12 = this.input;
			input12.bits_in_buffer -= 16;
			if (n8 != (this.uncomprLen ^ 0xFFFF)) {
				throw new DataFormatException("broken uncompressed block");
			}
			this.mode = 5;
		}
		case 5: {
			this.uncomprLen -= this.outputWindow.copyStored(this.input, this.uncomprLen);
			if (this.uncomprLen == 0) {
				this.mode = 2;
				return true;
			}
			return !this.input.needsInput();
		}
		case 6: {
			if (!this.dynHeader.decode(this.input)) {
				return false;
			}
			this.litlenTree = this.dynHeader.buildLitLenTree();
			this.distTree = this.dynHeader.buildDistTree();
			this.mode = 7;
			return this.decodeHuffman();
		}
		case 7:
		case 8:
		case 9:
		case 10: {
			return this.decodeHuffman();
		}
		case 12: {
			return false;
		}
		default: {
			throw new IllegalStateException();
		}
		}
	}

	static {
		CPLENS = new int[] { 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31, 35, 43, 51, 59, 67, 83, 99, 115,
				131, 163, 195, 227, 258 };
		CPLEXT = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0 };
		CPDIST = new int[] { 1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513, 769, 1025, 1537,
				2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577 };
		CPDEXT = new int[] { 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12,
				13, 13 };
	}
}
